import asyncio
import sys

import typer
from stlog import setup

from smartjob.app.executor import SmartJobExecutionResultFuture, SmartJobExecutorService
from smartjob.app.job import SmartJob, SmartJobExecutionResult
from smartjob.infra.controllers.lib import get_smart_job_executor_service_singleton


def get_job_service(ctx: typer.Context, **kwargs) -> SmartJobExecutorService:
    return get_smart_job_executor_service_singleton(
        namespace=ctx.obj.namespace,
        project=ctx.obj.project,
        region=ctx.obj.region,
        **kwargs,
    )


def init_stlog(ctx: typer.Context):
    setup(level=ctx.obj.log_level)


NameArgument = typer.Argument(help="Name of the job")
DockerImageArgument = typer.Option("", help="Docker image to use")
OverrideCommandArgument = typer.Option(
    "",
    help="Override docker image command and arguments",
)
OverrideEnvArgument = typer.Option(
    [],
    help="Override docker image env vars (key=value, can be used multiple times)",
)
StagingBucketArgument = typer.Option(
    None,
    envvar="SMARTJOB_STAGING_BUCKET",
    help="staging bucket (starting with gs://) for loading python_script_path",
)
InputBucketBasePathArgument = typer.Option(
    None, envvar="SMARTJOB_INPUT_BUCKET_BASE_PATH"
)
OutputBucketBasePathArgument: str = typer.Option(
    None, envvar="SMARTJOB_OUTPUT_BUCKET_BASE_PATH"
)
InputBucketPathArgument = typer.Option("")
OutputBucketPathArgument = typer.Option("")
PythonScriptPathArgument = typer.Option(
    "", help="local path to python script to execute in the container"
)
WaitArgument = typer.Option(True, help="Wait for the job to finish")


async def _run(
    service: SmartJobExecutorService, job: SmartJob
) -> SmartJobExecutionResult:
    return await service.run(job)


async def _schedule(
    service: SmartJobExecutorService, job: SmartJob
) -> SmartJobExecutionResultFuture:
    return await service.schedule(job)


def cli_process(service: SmartJobExecutorService, job: SmartJob, wait: bool):
    if wait:
        result = asyncio.run(_run(service, job))
        return_code = 0
        if result:
            print("SUCCESS in %i seconds" % (result.duration_seconds or -1))
        else:
            print("FAILED in %i seconds" % (result.duration_seconds or -1))
            return_code = 2
        print("Id:      %s" % result.job.execution_id)
        print("ShortId: %s" % result.job.short_execution_id)
        print("Logs:    %s" % result.log_url)
        sys.exit(return_code)
    else:
        future = asyncio.run(_schedule(service, job))
        print("SCHEDULED")
        print("Id:      %s" % future.job.execution_id)
        print("ShortId: %s" % future.job.short_execution_id)
        print("Logs:    %s" % future.log_url)



from abc import ABC, abstractmethod


class FileUploaderPort(ABC):
    @abstractmethod
    async def upload(
        self,
        content: str,
        destination_bucket: str,
        destination_path: str,
        only_if_not_exists: bool = True,
    ) -> str:
        pass



import shlex

import typer

from smartjob.app.job import CloudRunSmartJob
from smartjob.infra.controllers.cli.utils import (
    DockerImageArgument,
    InputBucketBasePathArgument,
    InputBucketPathArgument,
    NameArgument,
    OutputBucketBasePathArgument,
    OutputBucketPathArgument,
    OverrideCommandArgument,
    OverrideEnvArgument,
    PythonScriptPathArgument,
    StagingBucketArgument,
    WaitArgument,
    cli_process,
    get_job_service,
    init_stlog,
)

cli = typer.Typer()


@cli.command()
def run(
    ctx: typer.Context,
    name: str = NameArgument,
    docker_image: str = DockerImageArgument,
    override_command_and_args: str = OverrideCommandArgument,
    override_env: list[str] = OverrideEnvArgument,
    staging_bucket: str = StagingBucketArgument,
    input_bucket_base_path: str = InputBucketBasePathArgument,
    output_bucket_base_path: str = OutputBucketBasePathArgument,
    input_bucket_path: str = InputBucketPathArgument,
    output_bucket_path: str = OutputBucketPathArgument,
    python_script_path: str = PythonScriptPathArgument,
    wait: bool = WaitArgument,
    cpu: float = typer.Option(1.0, help="Number of CPUs"),
    memory_gb: float = typer.Option(0.5, help="Memory in Gb"),
):
    init_stlog(ctx)
    overriden_args = shlex.split(override_command_and_args)
    overriden_envs = {
        x.split("=")[0].strip().upper(): x.split("=")[1].strip() for x in override_env
    }
    service = get_job_service(
        ctx,
        input_bucket_base_path=input_bucket_base_path,
        output_bucket_base_path=output_bucket_base_path,
    )
    job = CloudRunSmartJob(
        name=name,
        docker_image=docker_image,
        overridden_args=overriden_args,
        overridden_envs=overriden_envs,
        staging_bucket=staging_bucket,
        input_bucket_path=input_bucket_path,
        output_bucket_path=output_bucket_path,
        python_script_path=python_script_path,
        cpu=cpu,
        memory_gb=memory_gb,
    )
    cli_process(service, job, wait)



from typing import Any

from smartjob.app.executor import SmartJobExecutorService
from smartjob.infra.adapters.executor.cloudrun import CloudRunSmartJobExecutor
from smartjob.infra.adapters.executor.vertex import VertexSmartJobExecutor
from smartjob.infra.adapters.file_uploader.gcs import GcsFileUploaderAdapter

# Default max workers for vertex executor and for file uploader (cloud run executor is not multi-threaded)
DEFAULT_MAX_WORKERS = 10
__singleton: SmartJobExecutorService | None = None


def get_smart_job_executor_service_singleton(
    max_workers: int = DEFAULT_MAX_WORKERS,
    namespace: str | None = None,
    project: str | None = None,
    region: str | None = None,
    staging_bucket: str | None = None,
    docker_image: str | None = None,
    input_bucket_base_path: str | None = None,
    output_bucket_base_path: str | None = None,
) -> SmartJobExecutorService:
    """Return a singleton instance of SmartJobExecutorService (initialized on first call with given arguments).

    Most of the arguments are optional and can also be set via environment variables or at SmartJob level.

    See SmartJobExecutorService or SmartJob for more details on the arguments.

    Args:
        max_workers: Maximum number of workers for the vertex executor and for the file uploader.
        namespace: Default namespace to use.
        project: Default project to use.
        region: Default region to use.
        staging_bucket: Default staging bucket to use.
        docker_image: Default docker image to use.
        input_bucket_base_path: Default input bucket base path to use.
        output_bucket_base_path: Default output bucket base path to use.

    Returns:
        Instance of SmartJobExecutorService.

    """
    global __singleton
    if __singleton is None:
        kwargs: dict[str, Any] = {}
        if input_bucket_base_path:
            kwargs["input_bucket_base_path"] = input_bucket_base_path
        if output_bucket_base_path:
            kwargs["output_bucket_base_path"] = output_bucket_base_path
        if docker_image:
            kwargs["docker_image"] = docker_image
        if staging_bucket:
            kwargs["staging_bucket"] = staging_bucket
        if region:
            kwargs["region"] = region
        if project:
            kwargs["project"] = project
        if namespace:
            kwargs["namespace"] = namespace
        __singleton = SmartJobExecutorService(
            cloudrun_executor_adapter=CloudRunSmartJobExecutor(),
            vertex_executor_adapter=VertexSmartJobExecutor(max_workers=max_workers),
            file_uploader_adapter=GcsFileUploaderAdapter(max_workers=max_workers),
            **kwargs,
        )
    return __singleton



import datetime
import hashlib
import uuid
from dataclasses import asdict, dataclass, field

from smartjob.app.exception import SmartJobException

# DEFAULT_NAMESPACE is the default namespace for jobs.
DEFAULT_NAMESPACE = "default"

# DEFAULT_TIMEOUT_SECONDS is the default timeout in seconds for jobs.
DEFAULT_TIMEOUT_SECONDS = 3600


def _hex_hash(*args: str) -> str:
    return hashlib.sha1("\n".join(args).encode()).hexdigest()


def _unique_id() -> str:
    return str(uuid.uuid4()).replace("-", "")


@dataclass
class SmartJob:
    """Base class for smart jobs."""

    name: str
    """Name of the job."""

    namespace: str = DEFAULT_NAMESPACE
    """Namespace of the job."""

    project: str = ""
    """GCP project to use for executing the job.

    If not set here, the value will be forced by the SmartJobExecutorService object.
    """

    region: str = ""
    """GCP region where to execute the job.

    If not set here, the value will be forced by the SmartJobExecutorService object.
    """

    docker_image: str = ""
    """Docker image to use for the job.

    Example: `docker.io/python:3.12`.
    """

    overridden_envs: dict[str, str] = field(default_factory=dict)
    """Environnement variables to inject into the container.

    Note: `INPUT_PATH` and `OUTPUT_PATH` will be automatically injected (if relevant).
    """

    overridden_args: list[str] = field(default_factory=list)
    """Container arguments (including command) to use.

    If not set, the default container image arguments will be used.

    """

    python_script_path: str = ""
    staging_bucket: str = ""
    input_bucket_path: str = ""
    output_bucket_path: str = ""
    execution_id: str | None = None
    timeout_seconds: int = DEFAULT_TIMEOUT_SECONDS
    service_account: str | None = None
    created: datetime.datetime | None = field(default=None, init=False)
    log_url: str | None = field(default=None, init=False)
    cancelled: bool = field(default=False, init=False)

    def set_execution_id(self):
        """Set a random execution_id."""
        self.execution_id = _unique_id()

    @property
    def short_execution_id(self) -> str:
        """Return a short id."""
        if self.execution_id is None:
            raise SmartJobException("execution_id is not set")
        return self.execution_id[:8]

    def assert_is_ready(self):
        """Test if the job is ready to be run.

        It raises an exception if the job is not ready to be run.
        If it's ok, it does nothing.

        Raises:
            SmartJobException: If the job is not ready to be run.

        """
        if not self.name:
            raise SmartJobException("name is required for job")
        if not self.project:
            raise SmartJobException("project is required")
        if not self.region:
            raise SmartJobException("region is required")
        if not self.docker_image:
            raise SmartJobException("docker_image is required")
        if self.staging_bucket:
            if not self.staging_bucket.startswith("gs://"):
                raise SmartJobException("staging_bucket must start with gs://")
            if "/" in self.staging_bucket[5:]:
                raise SmartJobException(
                    "staging_bucket name must not contain / (except for the the gs:// prefix)"
                )
        elif self.python_script_path:
            raise SmartJobException("staging_bucket is required for python_script_path")

    def python_script_hash(self) -> str:
        """Return the hash of the python script.

        If the python_script_path is not set, it returns an empty string.

        """
        if not self.python_script_path:
            return ""
        with open(self.python_script_path) as f:
            return hashlib.sha1(f.read().encode()).hexdigest()

    @property
    def staging_bucket_name(self) -> str:
        """Return the name of the staging bucket without the gs:// prefix.

        If it does not start with gs://, it returns an empty string.

        """
        if not self.staging_bucket.startswith("gs://"):
            return ""
        return self.staging_bucket[5:]  # let's remove gs:// prefix

    @property
    def input_bucket_name(self) -> str:
        """Return the name of the input bucket without the gs:// prefix.

        If it does not start with gs://, it returns an empty string.

        """
        if not self.input_bucket_path.startswith("gs://"):
            return ""
        return self.input_bucket_path[5:].split("/")[0]

    @property
    def _input_path(self) -> str:
        try:
            return "/".join(self.input_bucket_path[5:].split("/")[1:])
        except Exception:
            return ""

    @property
    def _output_path(self) -> str:
        try:
            return "/".join(self.output_bucket_path[5:].split("/")[1:])
        except Exception:
            return ""

    @property
    def output_bucket_name(self) -> str:
        """Return the name of the output bucket without the gs:// prefix.

        If it does not start with gs://, it returns an empty string.

        """
        if not self.output_bucket_path.startswith("gs://"):
            return ""
        return self.output_bucket_path[5:].split("/")[0]

    @property
    def staging_mount_point(self) -> str:
        """Return the mount point for the staging bucket."""
        raise NotImplementedError("mount_point must be implemented in subclasses")

    @property
    def input_mount_point(self) -> str:
        """Return the mount point for the input bucket."""
        raise NotImplementedError("mount_point must be implemented in subclasses")

    @property
    def output_mount_point(self) -> str:
        """Return the mount point for the output bucket."""
        raise NotImplementedError("mount_point must be implemented in subclasses")

    @property
    def input_path(self) -> str:
        """Return the input full path."""
        return f"{self.input_mount_point}/{self._input_path}"

    @property
    def output_path(self) -> str:
        """Return the output full path."""
        return f"{self.output_mount_point}/{self._output_path}"

    @property
    def full_name(self) -> str:
        """Return the full name of the job (namespace + name)."""
        return f"{self.namespace}-{self.name}"

    def set_auto_input_bucket_path(self, input_bucket_base_path: str):
        if not input_bucket_base_path.startswith("gs://"):
            raise SmartJobException("input_bucket_base_path must start with gs://")
        if not self.execution_id:
            raise SmartJobException("execution_id is not set")
        self.input_bucket_path = f"{input_bucket_base_path}/{self.full_name}/{self.execution_id[0:2]}/{self.execution_id}/input"

    def set_auto_output_bucket_path(self, output_bucket_base_path: str):
        if not output_bucket_base_path.startswith("gs://"):
            raise SmartJobException("output_bucket_base_path must start with gs://")
        if not self.execution_id:
            raise SmartJobException("execution_id is not set")
        self.output_bucket_path = f"{output_bucket_base_path}/{self.full_name}/{self.execution_id[0:2]}/{self.execution_id}/output"


@dataclass
class CloudRunSmartJob(SmartJob):
    # cpu is the number of requested CPUs.
    cpu: float = 1

    # memory_gb is the requested memory in Gb.
    memory_gb: float = 0.5

    @property
    def job_hash(self) -> str:
        """Return a short stable job_hash.

        If the job_hash is the same, we don't have to create a new CloudRunJob instance
        (but only a new job execution).

        """
        args: list[str] = [self.name, self.project, self.region, self.docker_image]
        args += [self.namespace, self.staging_bucket, self.service_account or ""]
        args += [self.input_bucket_name, self.output_bucket_name]
        return _hex_hash(*args)[:10]

    @property
    def _parent_name(self) -> str:
        return f"projects/{self.project}/locations/{self.region}"

    @property
    def cloud_run_job_name(self) -> str:
        """Return the cloud run job name."""
        return f"{self.namespace}-{self.name}-{self.job_hash}"

    @property
    def full_cloud_run_job_name(self) -> str:
        """Return the full cloud run job name."""
        return f"{self._parent_name}/jobs/{self.cloud_run_job_name}"

    @property
    def _cpuLimit(self) -> str:
        return "%f" % self.cpu

    @property
    def _memoryLimit(self) -> str:
        return "%f" % self.memory_gb + "Gi"

    @property
    def staging_mount_point(self) -> str:
        """Return the mount point for the staging bucket."""
        return "/staging"

    @property
    def input_mount_point(self) -> str:
        """Return the mount point for the input bucket."""
        return "/input"

    @property
    def output_mount_point(self) -> str:
        """Return the mount point for the output bucket."""
        return "/output"


@dataclass
class VertexSmartJob(SmartJob):
    machine_type: str = "n1-standard-4"
    accelerator_type: str = "ACCELERATOR_TYPE_UNSPECIFIED"
    accelerator_count: int = 0
    boot_disk_type: str = "pd-ssd"
    boot_disk_size_gb: int = 100

    @property
    def staging_mount_point(self) -> str:
        """Return the mount point for the staging bucket."""
        return f"/gcs/{self.staging_bucket_name}"

    @property
    def input_mount_point(self) -> str:
        """Return the mount point for the input bucket."""
        return f"/gcs/{self.input_bucket_name}"

    @property
    def output_mount_point(self) -> str:
        """Return the mount point for the output bucket."""
        return f"/gcs/{self.output_bucket_name}"

    def assert_is_ready(self):
        """Test if the job is ready to be run.

        It raises an exception if the job is not ready to be run.
        If it's ok, it does nothing.

        Raises:
            SmartJobException: If the job is not ready to be run.

        """
        super().assert_is_ready()
        if not self.staging_bucket:
            raise SmartJobException("staging_bucket is required for vertex jobs")


class SmartJobExecutionResult:
    """SmartJobExecutionResult is the (final) result of a job execution.

    Attributes:
        job: The executed job.
        success: Whether the job has succeeded or not.
        stopped: The datetime when the job has stopped.

    """

    def __init__(
        self,
        job: SmartJob,
        success: bool | None = None,
        stopped: datetime.datetime | None = None,
    ):
        self.job: SmartJob = job
        self.success: bool | None = success
        self.stopped: datetime.datetime | None = stopped

    def __bool__(self) -> bool:
        return self.success or False

    def __str__(self) -> str:
        if self.success:
            state = "SUCCESS"
        elif self.success is False:
            state = "FAILURE"
        else:
            state = "UNKNOWN"
        return f"SmartJobExecutionResult(short_execution_id={self.job.short_execution_id}, job_name={self.job.name}, job_namespace={self.job.namespace}): {state} in {self.duration_seconds or -1} seconds"

    def asdict(self) -> dict:
        """Return this object as a dictionary."""
        return {
            "success": self.success,
            "created": self.created,
            "stopped": self.stopped,
            "duration_seconds": self.duration_seconds,
            "job": asdict(self.job),
        }

    @property
    def duration_seconds(self) -> int | None:
        """The duration of the job in seconds."""
        if self.stopped is None:
            return None
        return (self.stopped - self.created).seconds

    @property
    def created(self) -> datetime.datetime:
        """The datetime when the job has started."""
        if not self.job.created:
            raise SmartJobException("created is not set")
        return self.job.created

    @property
    def log_url(self) -> str:
        """The log url of the job."""
        if not self.job.log_url:
            raise SmartJobException("log_url is not set")
        return self.job.log_url



from smartjob.app.exception import SmartJobException
from smartjob.app.executor import SmartJobExecutionResultFuture, SmartJobExecutorService
from smartjob.app.job import (
    CloudRunSmartJob,
    SmartJob,
    SmartJobExecutionResult,
    VertexSmartJob,
)
from smartjob.infra.controllers.lib import get_smart_job_executor_service_singleton

__all__ = [
    "get_smart_job_executor_service_singleton",
    "SmartJobException",
    "CloudRunSmartJob",
    "SmartJob",
    "SmartJobExecutorService",
    "SmartJobExecutionResult",
    "SmartJobExecutionResultFuture",
    "VertexSmartJob",
]



import asyncio
import datetime
import hashlib
import os
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Coroutine

from stlog import LogContext, getLogger

from smartjob.app.exception import SmartJobException
from smartjob.app.file_uploader import FileUploaderPort
from smartjob.app.job import (
    DEFAULT_NAMESPACE,
    CloudRunSmartJob,
    SmartJob,
    SmartJobExecutionResult,
    VertexSmartJob,
)

logger = getLogger("smartjob.executor")


class SmartJobExecutionResultFuture(ABC):
    """Future-like object to get the result of a job execution.

    Attributes:
        job: The job that was scheduled.

    """

    def __init__(
        self,
        coroutine: Coroutine,
        job: SmartJob,
    ):
        self.job = job
        self.__result: SmartJobExecutionResult | None = None
        self.__task = asyncio.create_task(coroutine)
        self.__task.add_done_callback(self._task_done)

    async def result(self) -> SmartJobExecutionResult:
        """Wait and return the result of the job execution.

        This is a coroutine. You have to wait for it to complete with (for example):
        `res = await future.result()`

        Returns:
            The result of the job execution.

        """
        if self.__result is not None:
            return self.__result
        task_result = await self.__task
        res = self._get_result(task_result)
        if self.__result is not None:
            res.stopped = self.__result.stopped
        else:
            res.stopped = datetime.datetime.now(tz=datetime.timezone.utc)
        return res

    def done(self) -> bool:
        """Return True if the job execution is done."""
        return self.__task.done()

    @abstractmethod
    def _get_result(self, task_result) -> SmartJobExecutionResult:
        pass

    @abstractmethod
    def _get_result_from_future(self, future) -> SmartJobExecutionResult:
        pass

    def _task_done(self, future: asyncio.Future):
        self.__result = self._get_result_from_future(future)
        self.__result.stopped = datetime.datetime.now(tz=datetime.timezone.utc)
        if self.job.cancelled:
            return
        if self.__result.success:
            logger.info(
                "Smartjob execution succeeded",
                duration_seconds=self.__result.duration_seconds,
            )
        else:
            logger.warn(
                "Smartjob execution failed",
                log_url=self.__result.log_url,
                duration_seconds=self.__result.duration_seconds,
            )

    @property
    def log_url(self) -> str | None:
        """Return the URL to the logs of the job execution."""
        return self.job.log_url


class SmartJobExecutorPort(ABC):
    @abstractmethod
    async def schedule(self, job: SmartJob) -> SmartJobExecutionResultFuture:
        pass


@dataclass
class SmartJobExecutorService:
    cloudrun_executor_adapter: SmartJobExecutorPort
    vertex_executor_adapter: SmartJobExecutorPort
    file_uploader_adapter: FileUploaderPort
    namespace: str = field(
        default_factory=lambda: os.environ.get("SMARTJOB_NAMESPACE", DEFAULT_NAMESPACE)
    )
    project: str = field(default_factory=lambda: os.environ.get("SMARTJOB_PROJECT", ""))
    region: str = field(default_factory=lambda: os.environ.get("SMARTJOB_REGION", ""))
    docker_image: str = field(
        default_factory=lambda: os.environ.get("SMARTJOB_DOCKER_IMAGE", "")
    )
    staging_bucket: str = field(
        default_factory=lambda: os.environ.get("SMARTJOB_STAGING_BUCKET", "")
    )
    input_bucket_base_path: str = field(
        default_factory=lambda: os.environ.get("SMARTJOB_INPUT_BUCKET_BASE_PATH", "")
    )
    output_bucket_base_path: str = field(
        default_factory=lambda: os.environ.get("SMARTJOB_OUTPUT_BUCKET_BASE_PATH", "")
    )

    def _update_job_with_default_parameters(self, job: SmartJob):
        if not job.namespace:
            job.namespace = self.namespace
        if not job.project:
            job.project = self.project
        if not job.region:
            job.region = self.region
        if not job.docker_image:
            job.docker_image = self.docker_image
        if not job.staging_bucket:
            job.staging_bucket = self.staging_bucket
        if self.input_bucket_base_path and not job.input_bucket_path:
            job.set_auto_input_bucket_path(self.input_bucket_base_path)
        if self.output_bucket_base_path and not job.output_bucket_path:
            job.set_auto_output_bucket_path(self.output_bucket_base_path)

    def _update_job_with_input_output(self, job: SmartJob):
        if job.input_bucket_path:
            job.overridden_envs["INPUT_PATH"] = job.input_path
        if job.output_bucket_path:
            job.overridden_envs["OUTPUT_PATH"] = job.output_path

    def _update_some_job_properties(self, job: SmartJob):
        job.created = datetime.datetime.now(tz=datetime.timezone.utc)
        job.set_execution_id()

    async def _create_input_output_paths_if_needed(self, job: SmartJob):
        coroutines: list[Coroutine] = []
        if job.input_bucket_path:
            logger.info(
                "Creating input path gs://%s/%s/...",
                job.input_bucket_name,
                job._input_path,
            )
            coroutines.append(
                self.file_uploader_adapter.upload(
                    "", job.input_bucket_name, job._input_path + "/"
                )
            )
        if job.output_bucket_path:
            logger.info(
                "Creating output path gs://%s/%s/...",
                job.output_bucket_name,
                job._output_path,
            )
            coroutines.append(
                self.file_uploader_adapter.upload(
                    "", job.output_bucket_name, job._output_path + "/"
                )
            )
        await asyncio.gather(*coroutines, return_exceptions=True)
        logger.debug("Done creating input/output paths")

    async def _upload_python_script_if_needed_and_update_overridden_args(
        self, job: SmartJob
    ):
        if not job.python_script_path:
            return
        if not job.staging_bucket:
            raise SmartJobException("staging_bucket is required for python_script_path")
        with open(job.python_script_path) as f:
            content = f.read()
        sha = hashlib.sha1(content.encode()).hexdigest()
        destination_path = f"{job.full_name}/{sha[0:2]}/{sha}.py"
        logger.info(
            "Uploading python script (%s) to %s/%s...",
            job.python_script_path,
            job.staging_bucket,
            destination_path,
        )
        await self.file_uploader_adapter.upload(
            content, job.staging_bucket_name, destination_path
        )
        logger.debug(
            "Done uploading python script (%s) to %s/%s",
            job.python_script_path,
            job.staging_bucket,
            destination_path,
        )
        job.overridden_args = [
            "python",
            f"{job.staging_mount_point}/{destination_path}",
        ] + job.overridden_args

    async def schedule(self, job: SmartJob) -> SmartJobExecutionResultFuture:
        """Schedule a job and return a kind of future.

        Arguments:
            job: The job to run.

        Returns:
            The result of the job execution as a kind of future.

        """
        self._update_some_job_properties(job)
        self._update_job_with_default_parameters(job)
        self._update_job_with_input_output(job)
        job.assert_is_ready()
        LogContext.reset_context()
        LogContext.add(
            job_name=job.name,
            execution_short_id=job.short_execution_id,
            job_namespace=job.namespace,
            project=job.project,
            region=job.region,
        )
        logger.info("Starting a smartjob...")
        await self._create_input_output_paths_if_needed(job)
        await self._upload_python_script_if_needed_and_update_overridden_args(job)
        res: SmartJobExecutionResultFuture
        if isinstance(job, CloudRunSmartJob):
            res = await self.cloudrun_executor_adapter.schedule(job)
        elif isinstance(job, VertexSmartJob):
            res = await self.vertex_executor_adapter.schedule(job)
        else:
            raise SmartJobException("Unknown job type")
        return res

    async def run(self, job: SmartJob) -> SmartJobExecutionResult:
        """Schedule a job and wait for its completion.

        Arguments:
            job: The job to run.

        Returns:
            The result of the job execution.

        """
        future = await self.schedule(job)
        return await future.result()
